# This module contains rutines for rendering with OpenGL.
# It uses Blenders Python OpenGL wrapper.

# Importing modules
import Blender
import re
from Blender import BGL
from Blender.BGL import *
from Blender import Draw
from Blender import Image
from Blender.Mathutils import *
from rahat.bfunc import *

def indexToCol(index):
    """
    indexToCol(index): returns color values as a list (R,G,B,A). It converts index into color.
    index is an integer.
    """
    color = Buffer( GL_BYTE, [4])    
    color[0] = 255 - (((index + 1) & 0xFF0000) >> 16)
    color[1] = 255 - (((index + 1) & 0x00FF00) >> 8)
    color[2] = 255 - ((index + 1) & 0x0000FF)
    color[3] = 255 - (((index + 1) & 0xFF000000) >> 24) # alpha may not been used!
    return color

def colToIndex(color):
    """
    colToIndex(color): returns index (int). It converts color into index.
    color is a list of 4 integer values (R,G,B,A).
    """
    index = ((255 - color[0]) << 16) + ((255 - color[1]) << 8) + (255 - color[2]) + ((255 - color[3]) << 24) - 1
    return index
    

def matToBuf(matrix):
    """
    matToBuf(matrix): returns Buffer object: transforms Matrix 4x4 type to Buffer.
    This is intendet to be used with OpenGL. So the order of stored values differs 
    from default C. Values in the buffer m1,m2,m3,m4,m5,... are these values in Matrix
    [ m1 m5     ]
    [ m2 ...    ]
    [ m3    ... ]
    [ m4    m16 ]
    """
    buf = Buffer( GL_DOUBLE, [16])
    i = 0
    for x in range(0,4):
        for y in range(0,4):
            buf[i]=matrix[y][x]  
            i = i+1
    return buf

def bufToMat(buffer):
    """
    bufToMat(buffer): returns Matrix object: transforms buffer to Matrix 4x4 type.
    This is intendet to be used with OpenGL. So the order of stored values differs 
    from default C. Values in the buffer m1,m2,m3,m4,m5,... are these values in Matrix
    [ m1 m5     ]
    [ m2 ...    ]
    [ m3    ... ]
    [ m4    m16 ]
    """
    mat = Matrix([0.0,0.0,0.0,0.0],
                 [0.0,0.0,0.0,0.0],
                 [0.0,0.0,0.0,0.0],
                 [0.0,0.0,0.0,0.0])
    i = 0
    for x in range(0,4):
        for y in range(0,4):
            mat[y][x]=buffer[i]  
            i = i+1
    return mat
    
def initView(fov, near, far, width, height, translation, rotation):
    """
    initView(fov, near, far, width, height, translation, rotation): initializes camera.
    It configures viewport and the rotation, translation of camera. And also the
    perspective transformations.
    fov is field of view in y axis. float.
    near is closest visible distance in z axis. float.
    far is the farest visible distance in z axis. float.
    width is the width of the viewport. integer.
    height is the height of the viewport. integer.
    translation is a vector of camera centre. list of floats. (x,y,z)
    rotation is a list of floats (angle, vector_x, vector_y, vector_z).
        where angle is a rotation around given vector.
    """
    # DEBUG
    print "initView(",fov,",",near,",",far,",",width,",",height,",",translation,",",rotation,")"
    # END DEBUG
    # Set the viewport
    glViewport( 0, 0, width, height)
    # Translate and rotate the view
    glMatrixMode( GL_MODELVIEW)
    glLoadIdentity()
    # Now there is a Identity Matrix in stack, after these two operations
    # there will be Final Matrix = Identity * Rotation * Translation
    glRotatef( rotation[0], rotation[1], rotation[2], rotation[3])
    glTranslatef( translation[0], translation[1], translation[2])
    # Set the perspective transformations
    glMatrixMode( GL_PROJECTION)
    gluPerspective( fov, float(width)/float(height), near, far)

    
def initRender():
    """
    def initRender(): initializes the OpenGL in Blender. It turns off some 
    OpenGL capabilities used in Blender. They should be turned on after finishing
    operations. It also clears the buffers and the transformation matrices.
    """
    # DEBUG
    print "initRender()"
    # END DEBUG
    # Disable the scissor box
    glDisable( GL_SCISSOR_TEST)
    # Disable clip planes used by Blender
    clipPlanes = Buffer( GL_INT, 1)
    glGetIntegerv( GL_MAX_CLIP_PLANES, clipPlanes)
    for i in range(0, clipPlanes[0]):
            glDisable(GL_CLIP_PLANE0 + i)
    # Delete transformations
    glMatrixMode( GL_MODELVIEW)
    glLoadIdentity()
    # Delete the projective transformations
    glMatrixMode( GL_PROJECTION)
    glLoadIdentity()  
    # Set the clear color and clear the buffers
    glClearColor( 0,0,0,0)
    # Set the render mode
    glRenderMode( GL_RENDER)    
    # Clear the buffers
    glClear( GL_COLOR_BUFFER_BIT)
    # Set the z value
    glClearDepth(1.0)
    glClear( GL_DEPTH_BUFFER_BIT )
    # Enable depth buffer test
    glEnable( GL_DEPTH_TEST)
    glDepthFunc( GL_LEQUAL)
    # Set the polygon offset
    glPolygonOffset(0.0,0.0)
    print "Render mode prepared."

def renderObject(object, indexColor=1):
    """
    def renderObject(object, indexColor=1): renders a Blender Mesh object. It uses objects world
    transformation matrix. The viewport transformations should be configured before.
    """
    # DEBUG
    print "renderObject(",object,",",indexColor,")"
    # END DEBUG    
    if object.getType() != "Mesh" :
        print "error! def renderObject(object): given object is not a Mesh type"
        print "       object type is: " + object.getType()
        return None    
    # Set the model transformations
    glMatrixMode( GL_MODELVIEW)
    glPushMatrix() # store current settings
    # Blender has stored the transposed matrices probably for easier using
    # in OpenGL. We need to transpose it.
    glMultMatrixd( matToBuf( Matrix(object.getMatrix()).transpose()))
    # Render the polygons (faces)
    nmesh = getDeformedMeshData(object)   
    for face in nmesh.faces:
        # We use same color for every face      
        color = indexToCol( indexColor)       
        glColor4ubv( color)
        glBegin( GL_POLYGON)
        for ver in face.v:
            glVertex3f( ver[0], ver[1], ver[2])            
        glEnd()           
    # Finish drawing    
    glFinish() 
    glPopMatrix() # restore transformations
    print "Object: " + object.getName() + " was processed."


def indexRender(object):
    """
    def indexRender(object): renders a Blender Mesh object.
    The rendered faces of the Mesh have different colors. It can be used
    to determine which pixel belongs to which face with colToIndex(color)
    function.
    """
    # DEBUG
    print "indexRender(",object,")"
    # END DEBUG    
    if object.getType() != "Mesh" :
        print "error! def renderObject(object): given object is not a Mesh type"
        print "       object type is: " + object.getType()
        return None    
    # Set the model transformations
    glMatrixMode( GL_MODELVIEW)
    glPushMatrix() # store transformation settings
    # Blender has stored the transposed matrices probably for easier using
    # in OpenGL. We need to transpose it.
    glMultMatrixd( matToBuf( Matrix(object.getMatrix()).transpose()))
    # Render the polygons
    nmesh = getDeformedMeshData(object)
    faceid = 0
    for face in nmesh.faces:
        color = indexToCol( faceid)
        glColor4ubv( color)
        glBegin( GL_POLYGON)
        for ver in face.v:
            glVertex3f( ver[0], ver[1], ver[2])            
        glEnd()   
        faceid = faceid + 1
    # Finish drawing    
    glFinish() 
    glPopMatrix() # restore transformation settings
    print "Object: " + object.getName() + " was processed."

def getRenderData( width, height, createImage = True):
    """
    def getRenderData( width, height, createImage = True): returns a Buffer object.
    It is created from OpenGL color buffer. This buffer can be used for determining
    visible faces. Buffers size is [ height, width, 4]. 4 becouse of RGBA components.
    It also creates a Blender Image. You can use it to see render results in Blender. 
    So you can not use the image for retrieving visible faces.
    """    
    # DEBUG
    print "getRenderData(",width,",",height,",",createImage,")"
    # END DEBUG    
    # First we need a buffer to read into
    buffer = Buffer( GL_BYTE, [height, width, 4])   
    glReadPixels( 0, 0, width, height, GL_RGBA, GL_UNSIGNED_BYTE, buffer)
    # Now we create an Image
    if createImage:
        image = Image.New( "renderData", width, height, 24)    
        for x in range(0,width):
            for y in range(0,height):
                #index = colToIndex( buffer[y][x])
                #R = 255 * (index & 1 == 1)
                #G = 255 * (index & 2 == 2)
                #B = 255 * (index & 4 == 4)
                #A = 255
                R = buffer[y][x][0]
                G = buffer[y][x][1]
                B = buffer[y][x][2]
                A = buffer[y][x][3]
                image.setPixelI( x, y, (R,G,B,A))
    return buffer
   
def getDepthData( width, height):
    """
    def getDepthData( width, height): returns a Blender Buffer object with depth data.
    """    
    # DEBUG
    print "getDepthData(",width,",",height,")"
    # END DEBUG
    buffer = Buffer( GL_FLOAT, [height, width])   
    glReadPixels( 0, 0, width, height, GL_DEPTH_COMPONENT, GL_FLOAT, buffer)
    return buffer

def initFeedback(size=1024):
    """
    def initFeedback(size=1024): initializes feedback mode, returns feedback buffer
    size is size of feedback buffer.
    """
    # DEBUG
    print "initFeedback(",size,")"
    # END DEBUG    
    # Set select mode
    buffer = Buffer( GL_FLOAT, size) 
    glFeedbackBuffer( size, GL_3D, buffer)
    glRenderMode(GL_FEEDBACK)
    return buffer
    
def passThroughRender(object):
    """
    def passThroughRender(object): renders a Blender Mesh object in  OpenGL select mode
    """
    # DEBUG
    print "passThroughRender(",object,")"
    # END DEBUG
    if object.getType() != "Mesh" :
        print "error! def renderObject(object): given object is not a Mesh type"
        print "       object type is: " + object.getType()
        return None    
    # Clear the transformations
    glMatrixMode( GL_MODELVIEW)
    glPushMatrix() # store transformation settings
    # Set the model transformations    
    glMultMatrixd( matToBuf( Matrix(object.getMatrix()).transpose()))
    # Render the polygons
    nmesh = getDeformedMeshData(object)
    faceid = 0
    for face in nmesh.faces:
        color = indexToCol( faceid)
        glColor4ubv( color)
        # Put the pass trough token
        glPassThrough( float(faceid))
        glBegin( GL_POLYGON)
        for ver in face.v:
            glVertex3f( ver[0], ver[1], ver[2])            
        glEnd()   
        faceid = faceid + 1
    # Finish drawing    
    glFinish() 
    glPopMatrix() # restore transformation settings
    print "Object: " + object.getName() + " was processed."
    

def getVisibleFaces( object):
    """
    def getVisibleFaces( object): returns a list of visible faces
    of the Blender Mesh object. Returns a list of values
    [ [faceid, visiblePixels, hiddenPixels], ..]    
    the viewport and view transformations must be defined before
    """
    # DEBUG
    print "getVisibleFaces(",object,")"
    # END DEBUG
    if object.getType() != "Mesh" :
        print "error! def renderObject(object): given object is not a Mesh type"
        print "       object type is: " + object.getType()
        return None    
    # Get the viewport, note that OpenGL stores data in different order
    # so it contains [left,bottom,right,top] instead of [left,right,bottom,top]
    view = Buffer( GL_INT, [4]) 
    glGetIntegerv( GL_VIEWPORT, view)
    # Render the object with indexes
    indexRender(object)  
    # Get the rendered data and do not create a image
    #print "Viewport: ", view
    rData = getRenderData(view[2], view[3], False)
    # Get the UV values of vertices
    uv = getUV( object)
    #print "UV: ",uv
    # Set the feedback mode and render the object
    fBuf = initFeedback( 512 * len(getDeformedMeshData(object).faces) )
    passThroughRender( object)
    # Get the number of hits in feedback buffer
    size = glRenderMode( GL_RENDER)
    print "size: ",size
    # We will test all the visible polygons on the screen if they
    # are visible or not. We will count visible and invisible pixels on 
    # the screen. We display the polygons again and will test the pixels.
    
    # We set up the propriet OpenGL mode
    # The informations we get from feedback buffer are after
    # all transformations and these coordinates are window coordintes
    # So we need no projection or etc.
    glClear( GL_COLOR_BUFFER_BIT)
    glDisable( GL_DEPTH_TEST)
    glMatrixMode( GL_PROJECTION)
    glLoadIdentity()
    gluOrtho2D( view[0], view[2], view[1], view[3])        
    glMatrixMode( GL_MODELVIEW)
    glPushMatrix() # store settings
    glLoadIdentity()
    # Now we prepare all needed variables
    faces = list() # this is the list of visible faces
    bbox = [-1,-1,-1,-1] # bounding box of polygon left,right,bottom,top
    count = 0 # offset in buffer
    faceid = -1 # polygon id in mesh
    visible = 0 # number of visible pixels
    hidden = 0 # number of hidden pixels
    while count < size:
        token = fBuf[ count]
        count = count + 1
        if token == GL_PASS_THROUGH_TOKEN:
            # Before processing new polygon, we test the last on visibility
            if faceid != -1:
                #print "faceid:",faceid
                # We finish drawing and will read the data
                glFinish()
                width = bbox[1]-bbox[0]
                height = bbox[3]-bbox[2]
                #print "box: ",width,",",height
                # We read the data
                if (width != 0) and (height != 0):
                    testBuffer = Buffer( GL_BYTE, [height, width,4])   
                    glReadPixels( bbox[0], bbox[2], width, height, GL_RGBA, GL_UNSIGNED_BYTE, testBuffer)
                    # Now we test the pixels
                    for x in range(0,width):
                        for y in range(0,height):
                            indexT = colToIndex( testBuffer[y][x])                            
                            indexC = colToIndex( rData[ y + bbox[2]][ x + bbox[0]])                            
                            # If there should be visible pixel of the tested polygon
                            # and is not we count hidden
                            if (indexT != indexC) and (indexT == faceid):
                                hidden = hidden + 1
                            # If there should be and is visible, we count visible
                            if (indexT == indexC) and (indexT == faceid):
                                visible = visible + 1
                # finaly we append this face as visible
                if visible > 0:
                    #print "Visible:",faceid,",",visible,",",hidden
                    faces.append( [faceid,visible,hidden])
            # Clear all variables
            visible = 0
            hidden = 0
            # We determine face id from the token
            faceid = int( fBuf[ count])            
            count = count + 1    
            # Every face we render again so we need to set the face color
            # to determine the index later
            color = indexToCol( faceid)
            glColor4ubv( color)
            # And clear the bounding box
            bbox = [-1,-1,-1,-1]
        # This token specifies polygon that should be rendered
        elif token == GL_POLYGON_TOKEN:    
            # Number of vertices
            vertices = int( fBuf[ count]); count = count + 1
            #print "vertices:",vertices            
            # Render the polygon
            glBegin( GL_POLYGON)
            for i in range(0,vertices):
                # Get the coordinates of vertex                
                vertex = [ fBuf[ count], fBuf[ count+1], fBuf[ count+2]]
                count = count + 3                
                # And show
                glVertex3f( vertex[0], vertex[1], vertex[2])
                # Correct the bounding box                
                if (bbox[0] == -1) or (bbox[0] > vertex[0]):
                    bbox[0] = int( vertex[0])
                if (bbox[1] == -1) or (bbox[1] < vertex[0]):
                    bbox[1] = int( vertex[0])
                if (bbox[2] == -1) or (bbox[2] > vertex[1]):
                    bbox[2] = int( vertex[1])
                if (bbox[3] == -1) or (bbox[3] < vertex[1]):
                    bbox[3] = int( vertex[1])
            glEnd()
    #print "Visible faces: ", len(faces)
    glPopMatrix() # restore transformation settings
    return faces

def printFeedback(passBuffer):
    """
    printFeedback(passBuffer): parse and print opengl feedback buffer     
    """
    # DEBUG
    print "printFeedback(",passBuffer,")"
    # END DEBUG    
    size = glRenderMode( GL_RENDER)    
    count = 0 # pozice v bufferu
    while count < size:
        token = passBuffer[ count]
        count = count + 1
        if token == GL_PASS_THROUGH_TOKEN:
            print "GL_PASS_THROUGH_TOKEN"
            faceid = int( passBuffer[ count])
            print "faceid:", faceid
            count = count + 1    
        elif token == GL_POLYGON_TOKEN:    
            print "GL_POLYGON_TOKEN"
            vertices = int( passBuffer[ count]); count = count + 1
            print "vertices:",vertices            
            for i in range(0,vertices):
                vertex = [ passBuffer[ count], passBuffer[ count+1], passBuffer[ count+2]]
                count = count + 3                
                print "vertex:",vertex
  


def getUV(object):
    """
    def getUV(object): returns list of [[u1,v1],[u2,v2],..] UV coordinates
    of face vertices of the object. The viewing transformations must be set before.
    The OpenGL state is queried to determine actual viewport transformations.
    """
    # DEBUG
    print "getUV(",object,")"
    # END DEBUG    
    if object.getType() != "Mesh" :        
        print "error! def getUV(object): given object is not a Mesh type"
        print "       object type is: " + object.getType()
        return None  
    # Get the Projection matrix 4x4 
    projection = Buffer( GL_DOUBLE, [16]) 
    glGetDoublev( GL_PROJECTION_MATRIX, projection)
    # Get the viewport dimensions
    view = Buffer( GL_INT, [4]) 
    glGetIntegerv( GL_VIEWPORT, view)
    # Get the final model transformations        
    model = Buffer( GL_DOUBLE, [16]) 
    glGetDoublev( GL_MODELVIEW_MATRIX, model)  
    # We need a buffers for each value
    UV = list()
    wx = Buffer( GL_DOUBLE, 1)   # protoze gluProject vraci double
    wy = Buffer( GL_DOUBLE, 1)   
    wz = Buffer( GL_DOUBLE, 1)    
    # Get the faces and uv coordinates for vertices
    nmesh = getDeformedMeshData(object)   
    faceid = 0
    for face in nmesh.faces:   
        uvs = list()     
        for ver in face.v:
            gluProject( ver[0], ver[1], ver[2], model, projection, view, wx, wy, wz)            
            uvs.append( [wx[0],wy[0]] )
        UV.append( uvs)
        faceid = faceid + 1    
    return UV

def setProjection( projMat, width, height, near=0.1, far=100.0):
    """
    def setProjection( projMat, width, height, near=0.1, far=100.0): loads a projection matrix 
    and setups the closest and the farest Z distance from camera.
    near and far should be always positive
    projMat must be 4x4 Blender Matrix
    width and height is the size of the viewport
    """
    # DEBUG
    print "setProjection(",projMat,",",width,",",height,",",near,",",far,")"
    # END DEBUG
    # aspect ratio of viewport
    aspect=float(width)/float(height)
    # Set Viewport and clear
    glViewport( 0, 0, width, height)
    glMatrixMode( GL_PROJECTION)
    glLoadIdentity()
    P=Matrix([2.0/width, 0.0, 0.0, 0.0],
             [0.0, 2.0/height, 0.0 ,0.0],
             [0.0, 0.0, 2.0/(far-near), -(far+near)/(far-near)],
             [0.0, 0.0, 1.0, 0.0])
    glMultMatrixd( matToBuf( P))
    # We set the perspective with the field of view 90 degrees in y axis
    # it allows us to say, that the projection division y:z is
    # 1:1. We pass the aspect ratio to get same transformations in x and y
    #gluPerspective( 90.0, aspect, near, far)
    # Everything else we do in modelview, note this:
    # viz OpenGL redbook Appendix F (p.461), ortho matrix
    # and glSpec2.0 p.43 (57)
    # In ModelView
    # MultMatrix(M) works as NEW = CURRENT * M, C = current then
    # MultMatrix(M);MultMatrix(N);MultMatrix(O) will produce
    # ((CM)N)O
    # !! but in PROJECTION mode it is different - NEW = M*CURRENT
    glMatrixMode( GL_MODELVIEW)
    glLoadIdentity()     
    # This matrix transform camera coordinates to opengl   
    #PixToGL = Matrix([-2.0/height,0.0,1.0*aspect,0.0],
    #                 [0.0,2.0/height,-1.0,0.0],
    #                 [0.0,0.0,1.0,0.0],
    #                 [0.0,0.0,0.0,1.0])
    PixToGL = Matrix([0.0,1.0,-width/2,0.0],
                     [-1.0,0.0,height/2,0.0],
                     [0.0,0.0,1.0,0.0],
                     [0.0,0.0,0.0,1.0])
    glMultMatrixd( matToBuf( PixToGL))  
    glMultMatrixd( matToBuf( projMat))
    
def loadCamPMat( filename):
    """
    def loadCamPMat( filename): loads projection matrix from the file and returns it.
    It must be stored in ASCII as follows:
    -1232.2312, 1232.2313,  ..., ...;
    ...,            ...,    ..., ...;
    ...,            ...,    ..., ...;
    0.0,            0.0,    0.0, 1.0;
    the last row may be missing.
    """ 
    # DEBUG
    print "loadCamPMat(",filename,")"
    # END DEBUG    
    f=open( filename,'r')       
    P=Matrix([1.0,0,0,0],[0,1.0,0,0],[0,0,1.0,0],[0,0,0,1.0])   
    lines = f.readlines()
    row = 0
    for line in lines:
        nums = re.findall("(-*\d+\.*\d*)\s*",line) 
        for col in range(4):   
            print float(nums[col])         
            P[row][col]=float(nums[col])        
        row = row+1
    f.close()    
    return P

def testGL():
    """
    def testGL(): test Blenders GL capabilities
    """
    # DEBUG
    print "testGL()"
    # END DEBUG
    color = indexToCol( 0x01)       
    glColor4ubv( color)
    glBegin(GL_POLYGON)
    glVertex3f(0.0,0.0,0.0)
    glVertex3f(1.0,0.0,0.0)
    glVertex3f(1.0,1.0,0.0)
    glEnd()    
    color = indexToCol( 0x02)       
    glColor4ubv( color)
    glBegin(GL_POLYGON)
    glVertex3f(0.0,0.0,1.0)
    glVertex3f(0.0,1.0,1.0)
    glVertex3f(1.0,1.0,1.0)
    glEnd()    
    color = indexToCol( 0x03)       
    glColor4ubv( color)
    glBegin(GL_POLYGON)
    glVertex3f(-0.5,-0.5,-1.0)
    glVertex3f(-0.5,0.5,-1.0)
    glVertex3f(0.5,0.5,-1.0)
    glVertex3f(0.5,-0.5,-1.0)
    glEnd()
    color = indexToCol( 0x04)       
    glColor4ubv( color)
    glBegin(GL_POLYGON)
    glVertex3f(-0.75,-0.75,0.1)
    glVertex3f(-0.75,0.75,0.1)
    glVertex3f(0.75,0.75,0.1)
    glVertex3f(0.75,-0.75,0.1)
    glEnd()
    
    
def testGLView(width,height):
    """
    def testGLView(): test Blenders GL viewport capabilities
    """
    # DEBUG
    print "testGLView()"
    # END DEBUG
    glViewport(0,0,width,height)
    glMatrixMode(GL_MODELVIEW)
    glLoadIdentity()
    # rotace kvuli testu
    #glRotated(90.0,0.0,0.0,1.0)
    M = Matrix( [1.0,0.0,0.0,0.0],
                       [0.0,1.0,0.0,0.0],
                       [0.0,0.0,0.01,0.01],
                       [0.0,0.0,-1.0,1.0])
    #glLoadMatrixd( matToBuf(M))
    
def printMatrices():
    """
    def printMatrices(): prints the modelview and projection matrix on the screen
    """
    # DEBUG
    print "printMatrices()"
    # END DEBUG    
    buffer = Buffer( GL_DOUBLE, [16])
    glGetDoublev( GL_MODELVIEW_MATRIX, buffer)
    print "Modelview:"
    print bufToMat(buffer)    
    glGetDoublev( GL_PROJECTION_MATRIX, buffer)
    print "Projection:"
    print bufToMat(buffer)

    